/*********************************************************************
 * Program     : GMC ( Gnu Midi Compiler)                            *
 * Author      : Sohler Guenther                                     *
 * Version     : v1.0Alpha                                           *
 * Last Changed: 9-15-1999                                           *
 ********************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#define NOTESDEFINED	125
#define MIDIUNIT	384
#define MAXSEQUENCES	30
#define REFTUNE		60 

#define NOTEONEVENT	0
#define NOTEOFFEVENT	1
#define TEXTEVENT	2
#define TEMPOEVENT	3
#define MEASUREEVENT	4
#define OTHEREVENT	5

char const notename[NOTESDEFINED][20]={"C","C#","Cb","D","D#","Db","E","E#","Eb","F","F#","Fb",
                                 "G","G#","Gb","A","A#","Ab","B","B#","Bb",
                                 "c","c#","cb","d","d#","db","e","e#","eb","f","f#","fb",
                                 "g","g#","gb","a","a#","ab","b","b#","bb",
                                 "^c","^c#","^cb","^d","^d#","^db","^e","^e#","^eb","^f","^f#","^fb",
                                 "^g","^g#","^gb","^a","^a#","^ab","^b","^b#","^bb","p",
                                 "kickdrum","sidestick","snaredrum","handclap","snaredrum2","lowtom2","closedhihat","lowtom","pedalhihat","midtom2","openhihat","midtom","hightom2","crashcymbal","hightom","ridecymbal","chinesecymbal","ridebell","tambourine","splashcymbal","cowbell","crashcymbal","vibraslap","ridecymbal2","highbongo","lowbongo","mutehighconga","mutelowconga","lowconga","hightimbale","lowtimbale","highangogo","lowangogo","cabasa","matacas","shorthiwhistle","loblowwhistle","shortguiro","longguiro","claves","highwoodblock","lowwoodblock","mutecuica","opencuica","mutetriangle","opentriangle","shaker","jinglebell","belltree","castanets","mutesurdo","opensurdo","slap","scratch1","scratch2","sticks","squareclick","metronomeclick","metronomebell","kickdrum2","highq" };
unsigned char notenum[NOTESDEFINED]={60,65,59,62,63,61,64,65,63,65,66,64,67,68,66,69,70,68,71,72,70,
                             72,73,71,74,75,73,76,77,75,77,78,76,79,80,78,81,82,80,83,84,82,
                             84,85,83,86,87,85,88,89,87,89,90,88,91,92,90,93,94,92,95,96,94,0,
			164,165,166,167,168,169,170,171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215 ,156,157,158,159,160,161,162,163,155};			    


int transposetable[12][7]= 
{{0,2,4,5,7,9,11},{1,3,4,6,8,10,11},{2,4,5,7,9,11,12},
{3,4,6,8,10,11,13},{4,5,7,9,11,12,14},{5,7,9,11,12,14,16},
{6,8,10,11,13,15,16},{7,9,11,12,14,16,17},{8,10,11,13,15,16,18},
{9,11,12,14,16,17,19}, {10,11,13,15,16,18,20},{11,12,14,16,17,19,21}};


typedef struct
{
 unsigned char voice;
 unsigned char volume;
 unsigned char pitch;
 unsigned char reverb;
} noteevent;


typedef struct
{
 char nominator;
 char denominator;
} measureevent;

typedef union
{
 noteevent n;
 char *o;
 char *t;
 double tempo;
 measureevent measure;

} midievents1;

typedef struct
{
 int time;
 unsigned char type;
 midievents1 e; 
} midievents;

typedef struct
{
 int eventanz;
 int maxtime;
 int mintime;
 midievents *event;
} ssequence;

int beatspertakt=4; 
int nested=0;

int sequenceanz=0;
char sequencename[MAXSEQUENCES][30];
ssequence sequencevec[MAXSEQUENCES];

char *mystrchr(char *text,char c)
{
 while(*text != 0 && *text != c) text++;
 if(*text == '\0') return NULL;
 return text;
}

char *mystrdup(char *text)
{
 char *ptr;
 if ( text == NULL) return NULL;
 if(*text == '\0') return NULL;
 ptr=(char *) malloc((strlen(text)+1)*sizeof(char));
 if(ptr == NULL) { printf("Too less memory!\n"); }
 memcpy(ptr,text,strlen(text)+1);
 return ptr;
}

char *convertcr(char *result)
{
 char buffer1[80];
 int buf1len;
 int i;
 buf1len=0;

 for(i=0;i<strlen(result);i++)
 {
  if(result[i] == '#')
  {
   buffer1[buf1len++]=0x0a;
  }
  else if(result[i] == '_')
  {
   buffer1[buf1len++]=' ';
  }
  else
  {
   buffer1[buf1len++]=result[i];
  }
 }
 buffer1[buf1len++]=0;
 free(result);
 result=mystrdup(buffer1);
 return result;
}

int split(char *fields[20],char *line)
{
 int fieldanz=0;
 char *line1;
 while(*line != '\0')
 {
  while(*line == ' ' || *line =='\n' || *line == '\r' || *line=='\t') line++;
  line1=line;
  if(*line == '\0') return fieldanz;
  while(!(*line1 == ' ' || *line1 =='\n' || *line1 == '\r' || *line1=='\t' || *line1 == '\0' )) line1++;
  fields[fieldanz]=malloc((line1-line+1)*sizeof(char));
  memcpy(fields[fieldanz],line,line1-line);
  fields[fieldanz++][line1-line]='\0';
  line=line1;
 }
 return fieldanz;
}

int dotranspose(int tune,int howmuch,int reftune,int offset)
{
 if(tune>= 0x80)
 {
  return tune;
 }
 while(howmuch >= 7) { howmuch-=7; tune+=12; }
 while(howmuch < 0) { howmuch+=7; tune-=12; }
 reftune=tune-reftune;
 reftune=(reftune+1200)%12;
 tune+=transposetable[reftune][howmuch]-reftune;
 return tune+offset;
}

char *gettextoption(char *item[30],int splitnum,char *key,char *def)
{
 char *result;
 char *equal;
 int i;
 for(i=0;i<splitnum;i++)
 {
  equal=mystrchr(item[i],'=');
  if(equal != NULL)
  {
   if(memcmp(key,item[i],strlen(key)) == 0)
   {
    result=mystrdup(equal+1);
    return result;
   }
  }
 }
 return mystrdup(def);
}

double getoption(char *item[30],int splitnum,char *key,double def)
{
 char *data=gettextoption(item,splitnum,key,NULL);
 double value;
 if(data == NULL)
 {
  return def;
 }
 sscanf(data,"%lf",&value);
 free(data);
 return value;
}
 
void add_event(ssequence *s,midievents *event)
{
 if(s->maxtime < event->time) s->maxtime=event->time;
 if(s->mintime > event->time) s->mintime=event->time;
 if(event->type == NOTEONEVENT || event->type == NOTEOFFEVENT)
 {
  if(event->e.n.pitch == 0)
  {
   return;
  }
 }
 s->eventanz++;
 s->event=(midievents *) realloc((void *) s->event,s->eventanz*sizeof(midievents));
 s->event[s->eventanz-1]=*event; 
 switch(event->type)
 {
  case NOTEONEVENT:
  case NOTEOFFEVENT:
                   break;
  case TEXTEVENT   :s->event[s->eventanz-1].e.t=mystrdup(event->e.t);
                    s->event[s->eventanz-1].type=TEXTEVENT;
                    break;
  case OTHEREVENT  :s->event[s->eventanz-1].e.o=mystrdup(event->e.o);
                    s->event[s->eventanz-1].type=OTHEREVENT;
                    break;
  case TEMPOEVENT  :break;
  case MEASUREEVENT:break;
  default:	    printf("Unknown event!\n");
                   break;
 
 }
}

long gotopos(char *item[30],int splitnum)
{
 int insertpos=0;
 char name[80];
 double f;
 if(splitnum != 3)
 {
  printf("Syntax: at <name of inserted sequence> <occurence>\n");
  exit(1);
 }
 strcpy(name,item[1]);
 sscanf(item[2],"%lf",&f);
 if(strcmp(name,"takt") == 0)
 {
  insertpos=f*MIDIUNIT*beatspertakt;
 }
 else
 {
  printf("Any else referenceposition as takt is not supported yet!\n");
  exit(2);
 }
 return insertpos;
}

char *gettext(char *text,int strophe)
{
 int k;
 char *ptr=text,*ptr1;
 char *result;
 if(strophe == 0) return NULL;
 if(strophe == -1) return mystrdup(text);
 if(text == NULL) return NULL;
 for(k=0;k<strophe-1;k++)
 {
  while(*ptr != '|' && *ptr != '\0') ptr++;
  ptr++;
 }
 if(*ptr == '\0') return NULL;
 ptr1=ptr;
 while(*ptr1 != '|' && *ptr1 != '\0') ptr1++;
 result=malloc((ptr1-ptr+1)*sizeof(char));
 memcpy(result,ptr,ptr1-ptr);
 result[ptr1-ptr]='\0';
 return result;
}

int place_sequence(ssequence *sequence,char *item[30],int splitnum,int insertpos,int strophe)
{
 int cutend=0,cutbeg=0;
 int valid=0;
 midievents event;
 int offset=0;
 char *usetextptr;
 double tempo1=120.0,tempo2=120.0;
 int sequencefound=-1,i=0,j=0,num=0,transpose=0;
 double timescale=1.0,duration=1.0;
 int intro=0,ending=1,reverb=0;
 double gain=1.0;
 int usevoice,voice=0,usevolume,usereverb;
 if(splitnum < 2)  {printf("SYNTAX: place < sequencename> \n"); exit(3);}
 sequencefound=-1;
 for(i=0;i<sequenceanz;i++)
 {
  if(strcmp(sequencename[i],item[1]) == 0) sequencefound=i;
 }
 if(sequencefound == -1)
 {
  printf("Did not find sequence named %s\n",item[1]);
  exit(4);
 }
 num=getoption(item,splitnum,"num",1);
 transpose=getoption(item,splitnum,"transpose",0);
 timescale=getoption(item,splitnum,"timescale",1.0);
 duration=getoption(item,splitnum,"duration",1.0);
 voice=getoption(item,splitnum,"voice",-1);
 gain=getoption(item,splitnum,"gain",1);
 intro=getoption(item,splitnum,"intro",0);
 ending=getoption(item,splitnum,"ending",0);
 offset=getoption(item,splitnum,"offset",0);
 reverb=getoption(item,splitnum,"reverb",-1);
 cutend=getoption(item,splitnum,"cutend",-1);
 cutbeg=getoption(item,splitnum,"cutbeg",1<<30);
 tempo1=getoption(item,splitnum,"tempo1",-1);
 tempo2=getoption(item,splitnum,"tempo2",-1);
 if(cutend == -1) cutend=sequencevec[sequencefound].maxtime*num; else cutend*=MIDIUNIT;
 if(cutbeg == 1<<30)
 {
  cutbeg=sequencevec[sequencefound].mintime;
  if(cutbeg >0) cutbeg=0;
 } 
 else cutbeg*=MIDIUNIT; for(j=0;j<num;j++)

 {
  for(i=0;i<sequencevec[sequencefound].eventanz;i++)
  {
   event=sequencevec[sequencefound].event[i];
   valid=1;
   switch(event.type)
   {
    case NOTEONEVENT:
    case NOTEOFFEVENT:
                     event.e.n.pitch=dotranspose(event.e.n.pitch,transpose,REFTUNE,offset);
                     usevoice=sequencevec[sequencefound].event[i].e.n.voice;
                     usereverb=sequencevec[sequencefound].event[i].e.n.reverb;
                     if(voice != -1) usevoice=voice;
                     if(reverb != -1) usereverb=reverb;
                     event.e.n.voice=usevoice;
                     usevolume=event.e.n.volume*gain;
                     if(intro == 1) usevolume *= (event.time-cutbeg)/(double) (cutend-cutbeg);
                     if(ending == 1) usevolume *= (1-(event.time-cutbeg)/(double) (cutend-cutbeg));
                     if(usevolume > 127) usevolume=127;
                     event.e.n.volume=usevolume;                     
                     event.e.n.reverb=usereverb;
                     break;
    case TEXTEVENT:  
                     usetextptr=gettext(sequencevec[sequencefound].event[i].e.t,strophe);
                     event.e.t=mystrdup(usetextptr); 
                     valid=0;
                     if(event.e.t != NULL)
                      if(*event.e.t != '\0')
                       valid=1;
                     break;
    case TEMPOEVENT: break;
    case MEASUREEVENT: break;
    default:	printf("Unknown Event!\n");
                exit(5);
                break;
   }
   event.time=event.time*timescale+insertpos+j*sequencevec[sequencefound].maxtime*timescale;
   if(event.time-insertpos >= cutend && event.type != NOTEOFFEVENT) valid=0;
   if(event.time-insertpos > cutend && event.type == NOTEOFFEVENT) valid=0;
   if(event.time-insertpos < cutbeg ) valid=0;
   if(cutbeg > 0) event.time-=cutbeg;
   if(valid) add_event(sequence,&event);
  }
 }
 if(cutbeg < 0) cutbeg=0;
 if(tempo1 != -1 && tempo2 == 1)
 {
  tempo2=tempo1;
 }
 if(tempo1 != -1)
 {
  for(i=0;i<10;i++)
  {
   event.time=insertpos+cutbeg+(cutend-cutbeg)*i/10;
   event.type=TEMPOEVENT;
   event.e.tempo=tempo1*(10-i)/10+tempo2*i/10;
   add_event(sequence,&event);
  }
 }
 return cutend-cutbeg;
}

void recordsequence(ssequence *sequence,FILE *in)
{
 double tempo;
 char *text;
 int splitnum;
 int done;
 char *item[30];
 midievents event;
 int time=0;
 int nom,denom;
 double duration;
 int i,notefound,transpose,voice,offset;
 int volumeuse,voiceuse,volume,reverb;
 int strophe;
 char buffer[80],buffer1[80];
 int buf1len;
 char *tempbuf;
 double timescale;
 while(!feof(in))
 {
  if(fgets(buffer,80,in) != NULL)
  {
   splitnum=split(item,buffer);
   if(splitnum > 0 && *item[0] != '#')
   {
    notefound=-1;
    timescale=1.0;
    done=1;
    while(done)
    {
     done=0;
     if(*item[0] != '\0')
     {
      if(item[0][strlen(item[0])-1] == '/')
      {
       item[0][strlen(item[0])-1]='\0';
       timescale/=2.0;
        done=1;
      }
      if(item[0][strlen(item[0])-1] == '\\')
      {
       item[0][strlen(item[0])-1]='\0';
       timescale/=3.0;
        done=1;
      }
      if(item[0][strlen(item[0])-1] >='2' && item[0][strlen(item[0])-1] <= '9')
      {
       timescale*= (item[0][strlen(item[0])-1]-'0');
       item[0][strlen(item[0])-1]='\0';
      }
     }
    } 
    for(i=0;i<NOTESDEFINED;i++)
    {
     if(strcmp(notename[i],item[0]) == 0)
     {
      notefound=i;
     }
    }
    if(notefound != -1)
    {
     volumeuse=getoption(item,splitnum,"volume",64);
     duration=getoption(item,splitnum,"duration",1);
     voiceuse=getoption(item,splitnum,"voice",0);
     reverb=getoption(item,splitnum,"reverb",0);
     text=gettextoption(item,splitnum,"text",NULL);
     if(text != NULL)
     {
      buf1len=0;
      for(i=0;i<strlen(text);i++)
      {
       if(text[i] == '#')
       {
        buffer1[buf1len++]=0x0a;
       }
       else if(text[i] == '_')
       {
        buffer1[buf1len++]=' ';
       }
       else
       {
        buffer1[buf1len++]=text[i];
       }
      }
      buffer1[buf1len++]=0;
      free(text);
      text=mystrdup(buffer1);
     }
     if(text != NULL && *text != '\0')
     {
      event.time=time;
      event.type=TEXTEVENT;
      event.e.t=mystrdup(text);
      add_event(sequence,&event);
      free(text);
     }
     event.type=NOTEONEVENT;
     event.time=time;
     event.e.n.pitch=notenum[notefound];
     event.e.n.volume=volumeuse;
     event.e.n.voice=voiceuse;
     event.e.n.reverb=reverb;
     add_event(sequence,&event);
     event.type=NOTEOFFEVENT;
     event.time=time+MIDIUNIT*timescale*duration;
     add_event(sequence,&event);
     time+=MIDIUNIT*timescale;
    }
    else
    if(strcmp(item[0],"origin") == 0)
    {
     for(i=0;i<sequence->eventanz;i++) sequence->event[i].time-=time;
     sequence->maxtime=0;
     sequence->mintime-=time;
     time=0;
    }
    else
    if(strcmp(item[0],"tempo") == 0)
    {
     if(splitnum != 2) { printf("syntax: tempo < number> \n"); exit(6); }
     sscanf(item[1],"%lf",&tempo);
     event.type=TEMPOEVENT;
     event.time=time;
     event.e.tempo=tempo;
     add_event(sequence,&event);
    }
    else
    if(strcmp(item[0],"measure") == 0)
    {
     if(splitnum != 3) { printf("syntax: measure < numinator denominator > \n"); exit(6); }
     sscanf(item[1],"%d",&nom);
     sscanf(item[2],"%d",&denom);
     event.type=MEASUREEVENT;
     event.time=time;
     event.e.measure.nominator=nom;
     event.e.measure.denominator=denom;
     add_event(sequence,&event);
    }
    else
    if(strcmp(item[0],"end") == 0)
    {
     nested--;
     return;
    }
    else
    if(strcmp(item[0],"define") == 0)
    {
     if(splitnum < 2) { printf("SYNTAX: define < sequencename >\n"); exit(7); }
     if(nested != 0)
     {
      printf("Cannot do define within define between %s and %s\n",sequencename[sequenceanz-1],item[1]);
      exit(1);
     }
     nested++;
     strcpy(sequencename[sequenceanz],item[1]);
     sequencevec[sequenceanz].eventanz=0;
     sequencevec[sequenceanz].event=NULL;
     sequencevec[sequenceanz].maxtime=0;
     sequencevec[sequenceanz].mintime=1<<30;
     sequenceanz++;    
     recordsequence(&sequencevec[sequenceanz-1],in);
     transpose=getoption(item,splitnum,"transpose",0);
     voice=getoption(item,splitnum,"voice",-1);
     timescale=getoption(item,splitnum,"timescale",1.0);
     duration=getoption(item,splitnum,"duration",1.0);
     volume=getoption(item,splitnum,"volume",-1);
     offset=getoption(item,splitnum,"offset",0);
     reverb=getoption(item,splitnum,"reverb",-1);
     for(i=0;i<sequencevec[sequenceanz-1].eventanz;i++)
     {
      switch(sequencevec[sequenceanz-1].event[i].type)
      {
       case NOTEONEVENT:
       case NOTEOFFEVENT:
                         sequencevec[sequenceanz-1].event[i].e.n.pitch=dotranspose(
                         sequencevec[sequenceanz-1].event[i].e.n.pitch,transpose,REFTUNE,offset);
                         if(voice != -1) sequencevec[sequenceanz-1].event[i].e.n.voice=voice;
                         if(volume != -1) sequencevec[sequenceanz-1].event[i].e.n.volume=volume;
                         if(reverb != -1) sequencevec[sequenceanz-1].event[i].e.n.reverb=reverb;
                         break;
       case TEMPOEVENT:	break;
       case MEASUREEVENT:break;
       case TEXTEVENT : break;
       default:		printf("Unknown Event type1!\n");
                        printf("%d\n",sequencevec[sequenceanz-1].event[i].type);
                        exit(8);
                        break;
      }
      sequencevec[sequenceanz-1].event[i].time*=timescale;
     } 
     sequencevec[sequenceanz-1].maxtime*=timescale;
     sequencevec[sequenceanz-1].mintime*=timescale;
    }
    else 
    if(strcmp(item[0],"place") == 0)
    {
     if(splitnum <2 ) { printf("SYNTAX: place < sequencename>\n"); exit(9); }
     strophe=getoption(item,splitnum,"strophe",-1);
     time+=place_sequence(sequence,item,splitnum,time,strophe);
    } 
    else
    if(strcmp(item[0],"at") == 0)
    {
     time=gotopos(item,splitnum);
    }
    else
    if(strcmp(item[0],"info") == 0)
    {
     if(splitnum < 2) { printf("Syntax: info <text>\n"); exit(10); }
     event.type=TEXTEVENT;
     event.time=time;
     tempbuf=mystrdup(item[1]);
     tempbuf=convertcr(tempbuf);
     sprintf(buffer,"%c%c%s",13,10,tempbuf);
     event.e.o=mystrdup(buffer);
     free(tempbuf);
     add_event(sequence,&event);
    }
    else
    {
     printf("Unable to recognize %s token!\n",item[0]);
     exit(11);
    }
   }
   for(i=0;i<splitnum;free(item[i++]));
  }
 }
}

void freesequence(ssequence *s)
{
 int i;
 for(i=0;i<s->eventanz;i++)
 {
  switch(s->event[i].type)
  {
   case NOTEONEVENT:
   case NOTEOFFEVENT:
                     break;
   case OTHEREVENT:  free(s->event[i].e.o);
                   break;
   case TEXTEVENT: free(s->event[i].e.t);
                  break;
   case TEMPOEVENT:break;
   case MEASUREEVENT:break;
   default:	  printf("unknown event type2!\n");
                  exit(1);
                  break;
  }
 }
}

void writebew(FILE *out,unsigned long int i)
{
 fprintf(out,"%c%c%c%c",(unsigned char)(i>>24)&0xff,(unsigned char)(i>>16)&0xff,(unsigned char)(i>>8)&0xff,(unsigned char)i&0xff);
}

void writebehw(FILE *out,unsigned short int i)
{
 fprintf(out,"%c%c",(i>>8)&0xff,i&0xff);
}

void midinum(FILE *out,unsigned long int num)
{
 if(num >=0x1000000) fputc(0x80+((num>>28)&0x7f),out);
 if(num >=0x20000) fputc(0x80+((num>>21)&0x7f),out);
 if(num >=0x4000) fputc(0x80+((num>>14)&0x7f),out);
 if(num >=0x80) fputc(0x80+((num>>7)&0x7f),out);
 if(num >=0x0) fputc(0x00+((num>>0)&0x7f),out);
}

void swap(midievents *a,int x,int y)
{
 midievents t;
 t=a[x];
 a[x]=a[y];
 a[y]=t;
}

void quicksort( midievents *a, int l, int r)
{
 int i,j,v;
 if(r > l)
 {
  v=2*a[r].time +((a[r].type!=NOTEOFFEVENT)?1:0);
  i=l-1;
  j=r;
  for(;;)
  {
   while(2*a[++i].time+((a[i].type!=NOTEOFFEVENT)?1:0)<v);
   while(2*a[--j].time+((a[j].type!=NOTEOFFEVENT)?1:0)>v);
   if(i>=j) break;
   swap(a,i,j);
  }
  swap(a,i,r);
  quicksort(a,l,i-1);
  quicksort(a,i+1,r);
 }
}


void writemidi(char *name,ssequence *sequence)
{
 int channelvoice[16];
 int channelreverb[16];
 int channelused[16];
 int channeluse;

 FILE *out;
 int done=0;
 int time=0;
 int i,j;
 int midilen,midilenpos;
 quicksort(sequence->event,0,sequence->eventanz-1);
 if(sequence->eventanz > 0) { time=sequence->event[0].time; } else { time=0; }
 if(time < 0) 
 {
  time=-(((-time)+MIDIUNIT*beatspertakt-1)/MIDIUNIT/beatspertakt)*MIDIUNIT*beatspertakt;
 }
 out=fopen(name,"wb");
 for(i=0;i<16;i++)
 {
  channelvoice[i]=-1;
  channelused[i]=0;
  channelreverb[i]=-1;
 }
 if(out == NULL)
 {
  printf("Cannot open %s for writing!\n",name);
  exit(1);
 }
 writebew(out,0x4d546864);
 writebew(out,6);
 writebehw(out,0);
 writebehw(out,1);
 writebehw(out,MIDIUNIT);
 writebew(out,0x4d54726b);
 midilenpos=ftell(out);
 writebew(out,0);
 fprintf(out,"%c%c%c%cmidicomp Track",0x00,0xff,0x03,0x0e);
 for(i=0;i<sequence->eventanz;i++)
 {
  midinum(out,sequence->event[i].time-time);
  time=sequence->event[i].time;
  if(sequence->event[i].type == NOTEONEVENT)
  {
   if((sequence->event[i].e.n.pitch&0x80) == 0)
   {
    channeluse=-1;
    for(j=0;j<16;j++)
    {
     if(channelvoice[j] == sequence->event[i].e.n.voice && channelreverb[j] == sequence->event[i].e.n.reverb && j != 9)
     {
      channeluse=j;
      j=16;
     }
    }
    if(channeluse == -1)
    {
     for(j=0;j<16;j++)
     {
      if(channelused[j] == 0 && channelvoice[j] == -1 && channelreverb[j] == -1 && j!= 9)
      {
       channeluse=j;
       channelvoice[j]=sequence->event[i].e.n.voice;
       channelreverb[j]=sequence->event[i].e.n.reverb;
       fprintf(out,"%c%c%c%c%c%c%c",0xc0+j,channelvoice[j],0,0xb0+j,0x5b,channelreverb[j],0);
       j=16;
      }
     }
    }
    if(channeluse == -1)
    {
     for(j=0;i<16;i++)
     {
      if(channelused[j] == 0 && j != 9)
      {
       channeluse=j;
       channelvoice[j]=sequence->event[i].e.n.voice;
       channelreverb[j]=sequence->event[i].e.n.reverb;
       fprintf(out,"%c%c%c%c%c%c%c",0xc0+j,channelvoice[j],0,0xb0+j,0x5b,channelreverb[j],0);
       j=16;
      }
     }
    }
    if(channeluse == -1)
    {
     printf("Was not able to map tune to channel ! \n");
     exit(1);
    }
   }
   else
   {
    channeluse=9;
    sequence->event[i].e.n.pitch&=0x7f;
   }
   fprintf(out,"%c%c%c",0x90+channeluse,sequence->event[i].e.n.pitch,sequence->event[i].e.n.volume);
  }
  else
  if(sequence->event[i].type == NOTEOFFEVENT)
  {
   channeluse=-1;
   for(j=0;j<16;j++)
   {
    if(channelvoice[j] == sequence->event[i].e.n.voice && channelreverb[j] == sequence->event[i].e.n.reverb)
    {
     channeluse=j;
    }
   }
   if(sequence->event[i].e.n.pitch&0x80)
   {
    channeluse=9;
    sequence->event[i].e.n.pitch&=0x7f;
   }
   if(j == -1) { printf("Error in program!\n"); exit(1); }
   fprintf(out,"%c%c%c",0x80+channeluse,sequence->event[i].e.n.pitch,sequence->event[i].e.n.volume);
  }
  else
  if(sequence->event[i].type == TEXTEVENT)
  {
   if(sequence->event[i].e.t != NULL)
   {
    fprintf(out,"%c%c%c%s",0xff,0x05,strlen(sequence->event[i].e.t),sequence->event[i].e.t);
   }
   else
   {
    fprintf(out,"%c%c%c",0xff,0x05,0x00);
   }
  }
  else
  if(sequence->event[i].type == TEMPOEVENT)
  {
   fprintf(out,"%c%c",0xff,0x51);
   writebew(out,0x03000000+(long)(60000000.0/sequence->event[i].e.tempo));
  }
  else
  if(sequence->event[i].type == MEASUREEVENT)
  {
   fprintf(out,"%c%c%c%c%c%c%c",0xff,0x58,0x04,sequence->event[i].e.measure.nominator,(char)(log(sequence->event[i].e.measure.denominator)/log(2.0)+0.1),0x18,0x08);
  }
  else
  if(sequence->event[i].type == OTHEREVENT)
  {
   fprintf(out,"%s",sequence->event[i].e.o);
  }
 }
 while(done == 1); 
 fprintf(out,"%c%c%c%c",0x00,0xff,0x2f,0x00);
 midilen=ftell(out);
 fseek(out,midilenpos,SEEK_SET);
 writebew(out,midilen-midilenpos-4);
 fclose(out);
}

int main(int argc,char *argv[])
{
 int i;
 FILE *in;
 ssequence sequence;
 char midioutfilename[80];
 sequence.eventanz=0;
 sequence.event=NULL;
 sequence.maxtime=0;
 sequence.mintime=1<<30;
 if(argc != 2)
 {
  printf("SYNTAX: %s < ascii file > \n",argv[0]);
  exit(1);
 }
 in=fopen(argv[1],"r");
 if(in == NULL)
 {
  printf("Cannot open file %s\n",argv[1]);
  exit(1);
 }
 recordsequence(&sequence,in);
 if(nested != 0)
 {
  printf("You forgot to an end somewhere!\n");
  exit(1);
 }
 fclose(in);
 sprintf(midioutfilename,"%s.mid",argv[1]);
 writemidi(midioutfilename,&sequence);
 freesequence(&sequence);
 for(i=0;i<sequenceanz;i++) freesequence(&sequencevec[i]);
 return 0;
} 
